/* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
/* Copyright (c) 2023 Brett Sheffield <bacs@librecast.net> */

#include <librecast_pvt.h>
#include <librecast/mdex.h>
#include <librecast/mtree.h>
#include <librecast/sync.h>
#include <sys/param.h>
#include <string.h>
#include <unistd.h>

static void * mdex_stack_pop(mdex_stack_t *stack)
{
	return (stack->ptr) ? stack->stack[--(stack->ptr)] : NULL;
}

static int mdex_stack_push(mdex_stack_t *stack, void *ptr)
{
	if (!stack->ptr || stack->ptr >= MDEX_STACK_SZ) {
		stack->len += MDEX_STACK_SZ;
		stack->stack = realloc(stack->stack, stack->len * sizeof(void *));
		if (!stack->stack) return -1;
	}
	stack->stack[(stack->ptr)++] = ptr;
	return 0;
}

int mdex_get(mdex_t *mdex, unsigned char *hash, size_t hashlen, mdex_entry_t *entry)
{
	mdex_idx_t *idx;
	int rc;

	if (!mdex || !hash || !hashlen) return (errno = EINVAL), -1;
	idx = mdex->root;
	if (!idx) return (errno = ENOENT), -1;
	while ((rc = memcmp(idx->hash, hash, hashlen))) {
		idx = (rc < 0) ? idx->clown : idx->joker;
		if (!idx) return (errno = ENOENT), -1;
	}
	if (entry && idx->entry) {
		if (idx->entry->type == MDEX_ALIAS) {
			if (!memcmp(hash, idx->entry->ptr.data, hashlen))
				return (errno = ELOOP), -1;
			memcpy(hash, idx->entry->ptr.data, hashlen);
			return mdex_get(mdex, hash, hashlen, entry);
		}
		*entry = *idx->entry;
	}
	return 0;
}

static int mdex_alloc(mdex_t *mdex)
{
	size_t pages;
	void *ptr;
	int err;

	if (!mdex->pagesz) return (errno = EINVAL), -1;
	if (!mdex->entries) mdex->entries++;
	pages = howmany(mdex->entries * sizeof(mdex_idx_t), mdex->pagesz);

	if (!pages) return (errno = EINVAL), -1;

	ptr = calloc(pages, mdex->pagesz);
	if (!ptr || mdex_stack_push(&mdex->alloc_stack, ptr)) goto exit_cleanup;
	if (!mdex->head) {
		mdex->head = ptr;
		mdex->root = mdex->head;
	}
	mdex->next = ptr;
	mdex->last = (mdex_idx_t *)((char *)ptr + mdex->entries * sizeof(mdex_idx_t));

	pages = howmany(mdex->entries * sizeof(mdex_entry_t), mdex->pagesz);
	ptr = calloc(pages, mdex->pagesz);
	if (!ptr || mdex_stack_push(&mdex->alloc_stack, ptr)) goto exit_cleanup;
	if (!mdex->head_entry) {
		mdex->head_entry = ptr;
		mdex->next_entry = mdex->head_entry;
	}
	mdex->next_entry = ptr;
	return 0;
exit_cleanup:
	err = errno;
	free(ptr);
	errno = err;
	return -1;
}

int mdex_put(mdex_t *mdex, unsigned char *hash, size_t hashlen, mdex_entry_t *entry)
{
	mdex_idx_t *new, *idx, **ptr;
	int rc;

	if (!mdex || !hash || !hashlen) return (errno = EINVAL), -1;
	new = mdex_stack_pop(&mdex->idx_stack);
	if (!new) {
		if (!mdex->next) {
			if (mdex_alloc(mdex) == -1) return -1;
		}
		new = mdex->next++;
		if (mdex->next >= mdex->last) mdex->next = NULL; /* force allocation */
	}
	idx = mdex->root;
	memcpy(new->hash, hash, hashlen);
	while ((rc = memcmp(idx->hash, hash, hashlen))) {
		ptr = (rc < 0) ? &idx->clown : &idx->joker;
		if (!*ptr) { *ptr = new; break; }
		idx = *ptr;
	}
	if (entry) {
		if (!rc && idx) new = idx; /* overwrite existing entry */
		new->entry = mdex_stack_pop(&mdex->entry_stack);
		if (!new->entry) new->entry = mdex->next_entry++;
		*new->entry = *entry;
		if ((new->entry->type & MDEX_FILE) && new->entry->file.file.name) {
			/* increment reference count for filename */
			new->entry->file.file.ref++;
		}
	}
	return 0;
}

int mdex_alias(mdex_t *mdex, unsigned char *alias, size_t aliaslen,
		unsigned char *hash, size_t hashlen)
{
	mdex_entry_t entry = { .type = MDEX_ALIAS };
	mdex_idx_t *idx;
	int rc;

	if (!mdex || !hash || !hashlen || !alias || !aliaslen)
		return (errno = EINVAL), -1;
	idx = mdex->root;
	if (!idx) return (errno = ENOENT), -1;
	while ((rc = memcmp(idx->hash, hash, hashlen))) {
		idx = (rc < 0) ? idx->clown : idx->joker;
		if (!idx) return (errno = ENOENT), -1;
	}
	entry.ptr.size = hashlen;
	entry.ptr.data = idx->hash;
	return mdex_put(mdex, alias, aliaslen, &entry);
}

int mdex_tree_hash(unsigned char *hash, size_t hashlen, mtree_t *mtree, size_t n)
{
	return hash_generic_key(hash, hashlen, mtree_nnode(mtree, 0), HASHSIZE,
			(unsigned char *)&n, sizeof(n));
}

static int mdex_tree(mdex_t *mdex, mtree_t *mtree)
{
	net_tree_t *data;
	unsigned char hash[HASHSIZE];
	mdex_entry_t entry = {0};
	size_t chunks, hashsz, off = 0, len;
	int rc = -1;

	entry.type = MDEX_PTR | MDEX_OTI;
	entry.ptr.size = HASHSIZE * mtree->nodes;
	hashsz = mtree->nodes * HASHSIZE;
	len = sizeof (net_tree_t) + hashsz;
	data = malloc(len);
	if (!data || mdex_stack_push(&mdex->alloc_stack, data)) {
		free(data);
		return -1;
	}
	data->size = htobe64(mtree->len);
	memcpy(data->tree, mtree->tree, hashsz);

	/* break the tree into chunks for indexing */
	chunks = howmany(len, (size_t)MTREE_CHUNKSIZE);
	for (size_t z = 0; z < chunks; z++) {
		mdex_tree_hash(hash, HASHSIZE, mtree, z);
		entry.ptr.size = MIN(MTREE_CHUNKSIZE, len);
		entry.ptr.data = (char *)data + off;
		rc = mdex_put(mdex, hash, HASHSIZE, &entry);
		if (rc == -1) break;
		len -= MTREE_CHUNKSIZE, off += MTREE_CHUNKSIZE;
	}
	return rc;
}

int mdex_add(mdex_t *mdex, void *data, size_t len, q_t *q, int flags)
{
	(void) flags; /* unused */
	mtree_t *mtree;
	unsigned char *hash;
	mdex_entry_t entry = {0};
	size_t min, max, off = 0;
	int err;

	mtree = malloc(sizeof (mtree_t));
	if (!mtree) return -1;
	if (mtree_init(mtree, len) == -1) goto exit_cleanup;
	if (mtree_build(mtree, data, q) == -1) goto exit_cleanup;

	/* index mtree */
	mdex_tree(mdex, mtree);

	/* write entries for each chunk */
	entry.type = MDEX_PTR;
	min = mtree_subtree_data_min(mtree->base, 0);
	max = min + mtree->chunks;
	for (size_t z = min; z < max; z++) {
		hash = mtree_nnode(mtree, z);
		entry.ptr.size = MIN(MTREE_CHUNKSIZE, len);
		entry.ptr.data = (char *)data + off;
		mdex_put(mdex, hash, HASHSIZE, &entry);
		len -= MTREE_CHUNKSIZE; off += MTREE_CHUNKSIZE;
	}
	mtree_free(mtree);
	free(mtree);
	return 0;
exit_cleanup:
	err = errno;
	mtree_free(mtree);
	free(mtree);
	errno = err;
	return -1;
}

int mdex_addfile(mdex_t *mdex, const char *path, q_t *q, int flags)
{
	struct stat sb = {0};
	char *data;
	mtree_t *mtree;
	unsigned char *hash;
	mdex_entry_t entry = {0};
	size_t len, min, max, sz = 0;
	int err;

	data = lc_mmapfile(path, &sz, PROT_READ, MAP_PRIVATE, 0, &sb);
	if (data == MAP_FAILED) return -1;
	mtree = malloc(sizeof (mtree_t));
	if (!mtree) goto exit_cleanup;
	if (mtree_init(mtree, sz) == -1) goto exit_cleanup;
	if (mtree_build(mtree, data, q) == -1) goto exit_cleanup;

	/* index mtree */
	mdex_tree(mdex, mtree);

	entry.file.file.name = strdup(path);
	if (mdex_stack_push(&mdex->alloc_stack, entry.file.file.name))
		goto exit_cleanup;

	/* write alias for root (hash filename) */
	if (flags & MDEX_ALIAS) {
		unsigned char alias[HASHSIZE];
		unsigned char root[HASHSIZE];
		if (mdex->basedir) {
			len = strlen(mdex->basedir);
			if (!memcmp(mdex->basedir, path, len)) path += len + 1;
		}
		hash_generic(alias, sizeof alias, (unsigned char *)path, strlen(path));
		mdex_tree_hash(root, HASHSIZE, mtree, 0);
		mdex_alias(mdex, alias, sizeof alias, root, HASHSIZE);
	}

	/* write entries for each chunk */
	entry.type = MDEX_FILE | (flags & MDEX_RAND);
	entry.file.off = 0;
	len = mtree->len;
	min = mtree_subtree_data_min(mtree->base, 0);
	max = min + mtree->chunks;
	for (size_t z = min; z < max; z++) {
		hash = mtree_nnode(mtree, z);
		entry.file.size = MIN(MTREE_CHUNKSIZE, len);
		mdex_put(mdex, hash, HASHSIZE, &entry);
		len -= MTREE_CHUNKSIZE;
		entry.file.off += MTREE_CHUNKSIZE;
	}
	mtree_free(mtree);
	free(mtree);
	munmap(data, sz);
	return 0;
exit_cleanup:
	err = errno;
	if (mtree) {
		mtree_free(mtree);
		free(mtree);
	}
	munmap(data, sz);
	errno = err;
	return -1;
}

int mdex_del(mdex_t *mdex, unsigned char *hash, size_t hashlen)
{
	mdex_idx_t *idx;
	mdex_idx_t **ptr = NULL;
	int rc;

	if (!mdex || !hash || !hashlen) return (errno = EINVAL), -1;
	idx = mdex->root;
	while ((rc = memcmp(idx->hash, hash, hashlen))) {
		ptr = (rc < 0) ? &idx->clown : &idx->joker;
		if (!*ptr) return (errno = ENOENT), -1;
		idx = *ptr;
	}
	if ((idx->entry->type & MDEX_FILE) && idx->entry->file.file.name) {
		/* decrement reference count for filename */
		if (--idx->entry->file.file.ref == 0) {
			free(idx->entry->file.file.name);
		}
	}
	/* push deleted idx + entry onto the stack */
	if (mdex_stack_push(&mdex->idx_stack, idx)) return -1;
	if (mdex_stack_push(&mdex->entry_stack, idx->entry)) return -1;
	if ((!idx->clown) && (!idx->joker)) {
		*ptr = NULL;       /* no children, just remove */
	}
	else if (idx->clown && !idx->joker) {
		*ptr = idx->clown; /* hoist left child */
	}
	else if (!idx->clown && idx->joker) {
		*ptr = idx->joker; /* hoist right child */
	}
	else {                     /* two children */
		/* find successor (smallest node in right subtree) */
		mdex_idx_t *s;
		for (s = idx->joker, ptr = &idx->joker; s && s->clown; ptr = &s->clown, s = s->clown);
		memcpy(idx->hash, s->hash, hashlen);
		idx->entry = s->entry;
		*ptr = s->joker; /* delete successor from tree */
	}
	return 0;
}

void mdex_basedir(mdex_t *mdex, char *basedir)
{
	mdex->basedir = strdup(basedir);
}

mdex_t *mdex_init(size_t entries)
{
	mdex_t *mdex = calloc(1, sizeof (mdex_t));
	if (mdex) {
		mdex->pagesz = (size_t)sysconf(_SC_PAGESIZE); /* cannot be < 1 */
		mdex->entries = entries;
		if (entries) mdex_alloc(mdex);
	}
	return mdex;
}

void mdex_free(mdex_t *mdex)
{
	void *ptr;
	while ((ptr = mdex_stack_pop(&mdex->alloc_stack))) free(ptr);
	free(mdex->alloc_stack.stack);
	free(mdex->entry_stack.stack);
	free(mdex->idx_stack.stack);
	free(mdex->basedir);
	free(mdex);
}
