/* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
/* Copyright (c) 2023 Brett Sheffield <bacs@librecast.net> */

#include <librecast_pvt.h>
#include <librecast/crypto.h>
#include <librecast/net.h>
#include <librecast/sync.h>
#include <librecast/mtree.h>
#include <librecast/mdex.h>
#include <string.h>
#include <sys/types.h>

struct syncjob {
	sem_t done;
	lc_ctx_t *lctx;
	lc_stat_t *stats;
	unsigned char *hash;
	unsigned char *ptr;
	ssize_t rc;
	size_t sz;
	int flags;
	int err;
};

/* wrapper for pthread_cleanup_push */
inline static void _lc_channel_part(lc_channel_t *chan)
{
	(void)lc_channel_part(chan);
}

/* return number of bits set in bitmap (Hamming Weight) */
static unsigned int hamm(unsigned char *map, size_t len)
{
	unsigned int c = 0;
	if (!map) return c;
	while (len--) for (char v = map[len]; v; c++) v &= v - 1;
	return c;
}

ssize_t lc_recvtree(lc_ctx_t *lctx, unsigned char *hash, mtree_t *tree, lc_stat_t *stats,
		lc_sync_options_t *opt, int flags)
{
	(void)opt, (void)flags; /* unused */
	rq_oti_t oti;
	rq_scheme_t scheme;
	net_tree_t *data;
	lc_socket_t *sock;
	lc_channel_t *chan;
	ssize_t rc = -1;
	uint64_t F;
	uint16_t T;
	int err = 0;

	/* join channel with FEC encoding. Trees are sent with OTI header */
	sock = lc_socket_new(lctx);
	if (!sock) goto err_errno;
	chan = lc_channel_hash(lctx, hash, HASHSIZE, LC_DEFAULT_FLAGS, LC_DEFAULT_PORT);
	if (!chan) goto err_socket_close;
	if (lc_channel_bind(sock, chan) == -1) { err = errno; goto err_channel_free; }
	lc_channel_coding_set(chan, lctx->coding | LC_CODE_FEC_RQ | LC_CODE_FEC_OTI);
	if (lc_channel_join(chan) == -1) { err = errno; goto err_channel_free; }

	/* peek at OTI header of first packet, allocate buffer */
	if ((rc = lc_channel_oti_peek(chan, &oti, &scheme)) == -1) { err = errno; goto err_channel_part; }
	F = rq_oti_F(oti);
	T = rq_oti_T(oti);
	if (T != LC_DEFAULT_RQ_T) { err= EBADMSG; rc = -1; goto err_channel_part; }
	data = malloc(F);
	if (!data) { err = errno; goto err_channel_part; }

	/* receive tree */
	rc = lc_channel_recv(chan, data, (size_t)F, 0);
	if (rc != -1) {
		if (stats) stats->byt_gross += rc;
		data->size = be64toh(data->size);
		if (mtree_init(tree, data->size) == 0) {
			size_t byt = F - sizeof(uint64_t);
			memcpy(tree->tree, data->tree, byt);
			size_t len = tree->nodes * HASHSIZE;
			size_t parts = howmany(len, MTREE_CHUNKSIZE);
			unsigned char root[HASHSIZE] = {0};
			uint8_t *ptr = (uint8_t *)tree->tree;
			F = len;
			for (size_t z = 1; z < parts; z++) {
				/* receive rest of multi-part tree */
				ptr += byt; len -= byt;
				mdex_tree_hash(root, HASHSIZE, tree, z);
				lc_channel_part(chan);
				lc_channel_free(chan);
				lc_socket_close(sock); sock = lc_socket_new(lctx); /* new socket */
				chan = lc_channel_hash(lctx, root, HASHSIZE, LC_DEFAULT_FLAGS, LC_DEFAULT_PORT);
				if (!chan) { err = errno; goto err_socket_close; }
				rc = lc_channel_bind(sock, chan);
				if (rc == -1) { err = errno; goto err_channel_free; }
				lc_channel_coding_set(chan, lctx->coding | LC_CODE_FEC_RQ | LC_CODE_FEC_OTI);
				if (lc_channel_join(chan) == -1) { err = errno; goto err_channel_free; }
				byt = MIN(len, MTREE_CHUNKSIZE);
				rc = lc_channel_recv(chan, ptr, byt, 0);
				if (rc == -1) { err = errno; break; }
			}
			if (rc != -1) {
				rc = mtree_verify(tree);
				if (rc == -1) err = ENOMSG;
			}
		}
		else err = errno, rc = -1;
		if (rc != -1) rc = (size_t)F;
	}
	else err = errno;

	/* clean up */
	free(data);
err_channel_part:
	lc_channel_part(chan);
err_channel_free:
	lc_channel_free(chan);
err_socket_close:
	lc_socket_close(sock);
err_errno:
	if (err) errno = err;
	return rc;
}

ssize_t lc_sendtree(lc_ctx_t *lctx, unsigned char *hash, const mtree_t *tree,
		lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)stats, (void)opt, (void)flags; /* unused */
	net_tree_t *data;
	lc_socket_t *sock;
	lc_channel_t *chan;
	ssize_t rc;
	size_t hashsz = tree->nodes * HASHSIZE;
	size_t sz = sizeof (net_tree_t) + hashsz;

	/* prepare tree for sending */
	data = malloc(sz);
	if (!data) return -1;
	pthread_cleanup_push(free, data);
	data->size = htobe64(tree->len);
	memcpy(data->tree, tree->tree, hashsz);

	/* send tree with FEC encoding */
	sock = lc_socket_new(lctx);
	if (!sock) { rc = -1; goto err_free_data; }
	pthread_cleanup_push((void (*)(void *))lc_socket_close, sock);
	chan = lc_channel_hash(lctx, hash, HASHSIZE, LC_DEFAULT_FLAGS, LC_DEFAULT_PORT);
	if (!chan) { rc = -1; goto err_socket_close; }
	pthread_cleanup_push((void (*)(void *))lc_channel_free, chan);
	if (lc_channel_bind(sock, chan) == -1) { rc = -1; goto err_channel_free; }
	if ((flags & NET_LOOPBACK) && lc_socket_loop(sock, 1) == -1) {
		rc = -1; goto err_channel_free;
	}
	lc_channel_coding_set(chan, lctx->coding | LC_CODE_FEC_RQ | LC_CODE_FEC_OTI);
	rc = lc_channel_send(chan, data, sz, 0);
	while (rc > 0) rc = lc_channel_send(chan, NULL, 0, 0);

	/* clean up */
err_channel_free:
	pthread_cleanup_pop(1); /* lc_channel_free */
err_socket_close:
	pthread_cleanup_pop(1); /* lc_socket_close */
err_free_data:
	pthread_cleanup_pop(1); /* free(data) */
	return rc;
}

ssize_t lc_recvchunk(lc_ctx_t *lctx, unsigned char *hash, void *data, const size_t len,
		lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)opt;
	lc_socket_t *sock;
	lc_channel_t *chan;
	ssize_t rc;

	sock = lc_socket_new(lctx);
	if (!sock) { rc = -1; goto err_exit; }
	pthread_cleanup_push((void (*)(void *))lc_socket_close, sock);
	chan = lc_channel_hash(lctx, hash, HASHSIZE, LC_DEFAULT_FLAGS, LC_DEFAULT_PORT);
	if (!chan) { rc = -1; goto err_socket_close; }
	pthread_cleanup_push((void (*)(void *))lc_channel_free, chan);
	if (lc_channel_bind(sock, chan) == -1) { rc = -1; goto err_channel_free; }
	lc_channel_coding_set(chan, lctx->coding | LC_CODE_FEC_RQ);
	if (lc_channel_join(chan) == -1) { rc = -1; goto err_channel_free; }
	pthread_cleanup_push((void (*)(void *))_lc_channel_part, chan);
	rc = lc_channel_recv(chan, data, len, flags);
	if (stats && rc > 0) stats->byt_gross += rc;
	pthread_cleanup_pop(1); /* lc_channel_part */
err_channel_free:
	pthread_cleanup_pop(1); /* lc_channel_free */
err_socket_close:
	pthread_cleanup_pop(1); /* lc_socket_close */
err_exit:
	return rc;
}

ssize_t lc_sendchunk(lc_ctx_t *lctx, unsigned char *hash, const void *data, const size_t len,
		lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)stats, (void)opt, (void)flags; /* unused */
	lc_socket_t *sock;
	lc_channel_t *chan;
	ssize_t rc;

	sock = lc_socket_new(lctx);
	if (!sock) { rc = -1; goto err_exit; }
	pthread_cleanup_push((void (*)(void *))lc_socket_close, sock);
	chan = lc_channel_hash(lctx, hash, HASHSIZE, LC_DEFAULT_FLAGS, LC_DEFAULT_PORT);
	if (!chan) { rc = -1; goto err_socket_close; }
	pthread_cleanup_push((void (*)(void *))lc_channel_free, chan);
	if (lc_channel_bind(sock, chan) == -1) { rc = -1; goto err_channel_free; }
	if ((flags & NET_LOOPBACK) && lc_socket_loop(sock, 1) == -1) {
		rc = -1; goto err_channel_free;
	}
	lc_channel_coding_set(chan, lctx->coding | LC_CODE_FEC_RQ);
	rc = lc_channel_send(chan, data, len, 0);
	while (rc > 0) rc = lc_channel_send(chan, NULL, 0, 0);
err_channel_free:
	pthread_cleanup_pop(1); /* lc_channel_free */
err_socket_close:
	pthread_cleanup_pop(1); /* lc_socket_close */
err_exit:
	return rc;
}

#ifdef HAVE_MLD
static void share_send_data(lc_share_t *share, struct in6_addr *grp, struct mdex_mem_s *mem,
		int coding)
{
	lc_socket_t *sock;
	lc_channel_t *chan;
	ssize_t rc;
	sock = lc_socket_new(share->lctx);
	if (!sock) return;
	chan = lc_channel_init_grp(share->lctx, grp, LC_DEFAULT_PORT);
	if (!chan) goto err_socket_close;
	if (share->ifx) lc_socket_bind(sock, share->ifx);
	if (lc_channel_bind(sock, chan) == -1) goto err_channel_free;
	if ((share->flags & LC_SHARE_LOOPBACK) && lc_socket_loop(sock, 1) == -1)
		goto err_channel_free;
	lc_channel_coding_set(chan, share->lctx->coding | coding);
	if (lc_channel_send(chan, mem->data, mem->size, 0) == -1)
		goto err_channel_free;
	assert(chan->rq);
	for (int i = 1; i < rq_KP(chan->rq) + RQ_OVERHEAD * 2; i++) {
		rc = lc_channel_send(chan, NULL, 0, 0);
		if (rc == -1) break;
	}
err_channel_free:
	lc_channel_free(chan);
err_socket_close:
	lc_socket_close(sock);
}

static void share_send_file(lc_share_t *share, struct in6_addr *grp, mdex_entry_t *entry, int coding)
{
	const struct mdex_file_s *file = &entry->file;
	struct mdex_mem_s mem = { .size = file->size };
	struct stat sbs = {0};
	char *smap = NULL;
	char *src = file->file.name;
	size_t sz = 0;
	if ((smap = lc_mmapfile(src, &sz, PROT_READ, MAP_PRIVATE, 0, &sbs)) == MAP_FAILED)
		return;
	mem.data = smap + file->off;
	share_send_data(share, grp, &mem, coding);
	munmap(smap, sz);
}

static void *share_send_thread(void *arg)
{
	lc_share_t *share = (lc_share_t *)arg;
	q_job_t job = {0};
	while (1) {
		pthread_testcancel();
		if (q_wait(&share->q, &job) == -1) continue;
		if (job.arg) {
			pthread_cleanup_push(free, job.arg);
			int coding = share->lctx->coding | LC_CODE_FEC_RQ;
			struct qentry_s *qe = (struct qentry_s *)job.arg;
			mdex_entry_t *entry = &qe->entry;
			struct in6_addr *grp = &qe->grp;
			if (entry->type & MDEX_RAND) {
				coding |= LC_CODE_FEC_RAND;
			}
			if (entry->type & MDEX_OTI) {
				coding |= LC_CODE_FEC_OTI;
			}
			if (entry->type & MDEX_PTR) {
				share_send_data(share, grp, &entry->ptr, coding);
			}
			else if (entry->type & MDEX_FILE) {
				share_send_file(share, grp, entry, coding);
			}
			pthread_cleanup_pop(1); /* free(entry) */
		}
	}
	return NULL;
}

/* when MLD detects a JOIN, check mdex for matching group, and queue for sending */
static void share_watch_callback(mld_watch_t *watch)
{
	lc_share_t *share = (lc_share_t *)mld_watch_arg(watch);
	mdex_entry_t entry = {0};
	unsigned char hash[HASHSIZE];
	struct in6_addr *grp = mld_watch_grp(watch);
	memcpy(hash, &(grp->s6_addr[2]), 14);
	if (!mdex_get(share->mdex, hash, 14, &entry)) {
		if (!q_search(&share->q, NULL, &entry)) {
			struct qentry_s *arg = malloc(sizeof (struct qentry_s));
			if (arg) {
				arg->grp = *grp;
				arg->entry = entry;
				q_push(&share->q, NULL, arg);
			}
		}
	}
}
#endif

lc_share_t * lc_share(lc_ctx_t *lctx, mdex_t *mdex, unsigned int ifx, lc_stat_t *stats,
		lc_sync_options_t *opt, int flags)
{
#ifdef HAVE_MLD
	(void)stats, (void)opt, (void)flags; /* unused */
	const int nthread_send = SYNC_SEND_THREADS;
	lc_share_t *share;
	int err = 0;

	/* allocate & initialize share */
	share = calloc(1, sizeof (lc_share_t));
	if (!share) return NULL;
	share->tsend = calloc(nthread_send, sizeof(pthread_t));
	if (!share->tsend) goto err_free_share;
	share->lctx = lctx;
	share->mdex = mdex;
	share->ifx = ifx;
	share->flags = flags;

	/* initialize MLD and watch callback */
	share->mld = mld_init(0);
	if (!share->mld) goto err_free_tsend;
	share->watch = mld_watch_add(share->mld, ifx, NULL, share_watch_callback, share, MLD_EVENT_JOIN);
	if (!share->watch) goto err_free_mld;
	if (!mld_start(share->mld)) goto err_free_mld;

	/* initialize send queue and start send thread(s) */
	if (q_init(&share->q) == -1) goto err_free_mld;
	err = q_pool_create(share->tsend, nthread_send, share_send_thread, share);
	share->nthread_send = nthread_send - err; /* number of threads actually created */
	if (!share->nthread_send) { err = errno; goto err_free_q; }
	return share;
err_free_q:
	(void)q_free(&share->q);
err_free_mld:
	free(share->mld);
err_free_tsend:
	free(share->tsend);
err_free_share:
	free(share);
	errno = err;
	return NULL;
#else
	(void)lctx; (void)mdex; (void)ifx; (void)stats; (void)opt; (void)flags; /* unused */
	return (errno = ENOTSUP), NULL;
#endif
}

void lc_unshare(lc_share_t *share)
{
#ifdef HAVE_MLD
	int err = errno;
	q_job_t job = {0};
	q_pool_destroy(share->tsend, share->nthread_send);
	free(share->tsend);
	while (!q_trywait(&share->q, &job)) free(job.arg);
	(void)q_free(&share->q);
	mld_stop(share->mld);
	mld_watch_del(share->watch);
	mld_free(share->mld);
	free(share);
	errno = err;
#else
	(void)share;
#endif
}


int lc_memsync(void *dst, void *src, const size_t n, q_t *q, lc_stat_t *stats,
		lc_sync_options_t *opt, int flags)
{
	(void)stats, (void)opt, (void)flags; /* unused */
	mtree_t stree = {0}, dtree = {0};
	unsigned char *map;
	int rc = -1, err = 0;

	if (!n) return (errno = EINVAL), -1;

	/* create mtree for src and dest */
	if (mtree_init(&stree, n) == -1) { err = errno; goto err_errno; }
	if (mtree_init(&dtree, n) == -1) { err = errno; goto err_free_stree; }

	/* build trees */
	if (mtree_build(&stree, src, q) == -1) { err = errno; goto err_free_dtree; }
	if (mtree_build(&dtree, dst, q) == -1) { err = errno; goto err_free_dtree; }

	/* diff trees */
	map = mtree_diff_subtree(&stree, &dtree, 0, 1);
	if (!map) { err = errno; goto err_free_dtree; }

	/* sync data */
	for (size_t z = 0; z < stree.chunks; z++) {
		if (isset(map, z)) {
			size_t off = z * MTREE_CHUNKSIZE;
			size_t len = MIN(MTREE_CHUNKSIZE, n - off);
			memcpy((char *)dst + off, (char *)src + off, len);
		}
	}

	/* clean up & return */
	free(map);
err_free_dtree:
	rc = 0;
	mtree_free(&dtree);
err_free_stree:
	mtree_free(&stree);
err_errno:
	if (err) errno = err;
	return rc;
}

void *lc_mmapfile(const char *pathname, size_t *len, int prot, int flags, off_t offset,
		struct stat *sb)
{
	char *map = MAP_FAILED;
	mode_t mode;
	int oflag;
	int fd;
	oflag = (prot & PROT_WRITE) ? O_RDWR | O_CREAT : O_RDONLY;
	mode = (sb->st_mode) ? sb->st_mode : 0600;
	if ((fd = open(pathname, oflag, mode)) == -1) {
		goto err_exit;
	}
	if (!sb->st_ino && fstat(fd, sb) == -1) {
		goto err_close_fd;
	}
	if (*len && ftruncate(fd, *len) == -1) {
		goto err_close_fd;
	}
	if (!*len) *len = sb->st_size;
	map = mmap(NULL, *len, prot, flags, fd, offset);
err_close_fd:
	close(fd);
err_exit:
	return map;
}

int lc_syncfilelocal(const char *dst, const char *src, q_t *q, lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)stats, (void)opt, (void)flags; /* unused */
	struct stat sbs = {0}, sbd = {0};
	char *smap, *dmap;
	size_t sz = 0;
	int rc = -1, err = 0;

	smap = lc_mmapfile(src, &sz, PROT_READ, MAP_PRIVATE, 0, &sbs);
	if (smap == MAP_FAILED) { err = errno; goto err_errno; }
	sbd.st_mode = sbs.st_mode;
	sz = sbs.st_size;
	dmap = lc_mmapfile(dst, &sz, PROT_READ|PROT_WRITE, MAP_SHARED, 0, &sbd);
	if (dmap == MAP_FAILED) { err = errno; goto err_free_smap; }
	if (chmod(dst, sbs.st_mode) == -1) { err = errno; goto err_free_smap; }
	if (sbs.st_size > sbd.st_size) {
		/* source is longer, copy extra bytes before syncing first part */
		size_t ext = sbs.st_size - sbd.st_size;
		memcpy(dmap + sbs.st_size - ext, smap + sbs.st_size - ext, ext);
	}
	rc = lc_memsync(dmap, smap, sz, q, stats, opt, flags);
	msync(dmap, sz, MS_SYNC); /* on OpenBSD an explicit msync() is required */
	munmap(dmap, sz);
err_free_smap:
	munmap(smap, sz);
err_errno:
	if (err) errno = err;
	return rc;
}

ssize_t lc_sync(lc_ctx_t *lctx, unsigned char *hash, void *data, const size_t len,
		q_t *q, lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)stats, (void)opt, (void)flags; /* unused */
	mtree_t stree = {0};
	mtree_t dtree = {0};
	unsigned char *map;
	ssize_t byt = 0, rc;
	if ((rc = lc_recvtree(lctx, hash, &stree, stats, opt, flags)) == -1) return -1;
	if ((rc = mtree_init(&dtree, len) == -1)) goto err_free_stree;
	if ((rc = mtree_build(&dtree, data, q) == -1)) goto err_free_dtree;
	map = mtree_diff_subtree(&stree, &dtree, 0, 1);
	if (map) {
		size_t min = mtree_subtree_data_min(stree.base, 0);
		unsigned char *ptr;
		for (size_t z = 0; z < stree.chunks; z++) {
			if (isset(map, z)) {
				size_t off = z * MTREE_CHUNKSIZE;
				size_t sz = MIN(MTREE_CHUNKSIZE, len - off);
				ptr = (unsigned char *)data + off;
				unsigned char *chunkhash = mtree_nnode(&stree, min + z);
				rc = lc_recvchunk(lctx, chunkhash, ptr, sz, stats, opt, flags);
				if (rc == -1) break;
				byt += rc;
			}
		}
		free(map);
		if (rc != -1) rc = byt;
	}
err_free_dtree:
	mtree_free(&dtree);
err_free_stree:
	mtree_free(&stree);
	return rc;
}

static void * sync_chunk(void *arg)
{
	struct syncjob *job = (struct syncjob *)arg;
	unsigned char hash[HASHSIZE];
	do {
		job->rc = lc_recvchunk(job->lctx, job->hash, job->ptr, job->sz, job->stats, NULL, job->flags);
		hash_generic(hash, sizeof hash, job->ptr, job->sz);
	}
	while (memcmp(hash, job->hash, HASHSIZE));
	sem_post(&job->done);
	return NULL;
}

inline static ssize_t syncfile(lc_ctx_t *lctx, unsigned char *map, mtree_t *stree, char **dst,
		size_t len, q_t *q, lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	(void)opt; /* unused */
	struct syncjob *job;
	ssize_t byt = 0, rc = 0;
	size_t min = mtree_subtree_data_min(stree->base, 0);
	int diffs = hamm(map, stree->chunks);
	int err = 0, j = 0;
	job = calloc(diffs, sizeof(struct syncjob));
	if (!job) return -1;
	/* queue jobs to sync each differing chunk */
	for (size_t z = 0; z < stree->chunks; z++) {
		if (isset(map, z)) {
			size_t off = z * MTREE_CHUNKSIZE;
			job[j].lctx = lctx;
			job[j].sz = MIN(MTREE_CHUNKSIZE, len - off);
			job[j].ptr = (unsigned char *)dst + off;
			job[j].hash = mtree_nnode(stree, min + z);
			job[j].stats = stats;
			job[j].flags = flags;
			sem_init(&job[j].done, 0, 0);
			q_push(q, sync_chunk, &job[j++]);
		}
	}
	/* reap jobs, gather stats */
	for (int i = 0; i < diffs; i++) {
		sem_wait(&job[i].done);
		sem_destroy(&job[i].done);
		if (job[i].rc > 0) {
			/* Gross rate set by lc_recvtree & lc_recvchunk */
			if (stats) stats->byt_info += job[i].sz; /* Information Rate */
			byt += job[i].sz;
		}
		else {
			rc = -1;
			err = job->err;
		}
	}
	if (rc != -1) rc = byt;
	else err = errno;
	free(job);
	if (err) errno = err;
	return rc;
}

ssize_t lc_syncfile(lc_ctx_t *lctx, unsigned char *hash, const char *pathname,
		q_t *q, lc_stat_t *stats, lc_sync_options_t *opt, int flags)
{
	mtree_t stree = {0}, dtree = {0};
	struct stat sbd = {0};
	char **dst = NULL;
	unsigned char *map;
	pthread_t *tid;
	ssize_t byt = 0, rc;
	size_t len = 0;
	int err = 0, qfree = 0, nthreads;

	if ((rc = lc_recvtree(lctx, hash, &stree, stats, opt, flags)) == -1) return -1;
	nthreads = MIN(MTREE_THREADMAX, stree.base);
	tid = calloc(nthreads, sizeof(pthread_t));
	if (!tid) {
		err = errno;
		rc = -1;
		goto err_free_stree;
	}
	if (!q) { /* no queue supplied, allocate one */
		if (!(q = calloc(1, sizeof *q))) { err = errno; goto err_free_tid; }
		qfree |= 1;
		if (q_init(q) == -1) { err = errno; goto err_free_q; }
		qfree |= 2;
		if (q_pool_create(tid, nthreads, q_job_seek, q) != 0) {
			err = errno; goto err_free_q;
		}
		qfree |= 4;
	}
	len = stree.len;
	dst = lc_mmapfile(pathname, &len, PROT_READ|PROT_WRITE, MAP_SHARED, 0, &sbd);
	if (dst == MAP_FAILED) {
		err = errno;
		rc = -1;
		goto err_free_q;
	}
	if (mtree_init(&dtree, len) == -1) {
		err = errno;
		rc = -1;
		goto err_free_dst;
	}
	if (mtree_build(&dtree, dst, q) == -1) {
		err = errno;
		rc = -1;
		goto err_free_dtree;
	}
	map = mtree_diff_subtree(&stree, &dtree, 0, 1);
	if (map) {
		rc = syncfile(lctx, map, &stree, dst, len, q, stats, opt, flags);
		if (rc != -1) rc = byt;
		else err = errno;
		msync(dst, len, MS_SYNC); /* on OpenBSD an explicit msync() is required */
		free(map);
	}
	/* Clean up. My goodness we've made such an awful mess. Sorry Mum! */
err_free_dtree:
	mtree_free(&dtree);
err_free_dst:
	munmap(dst, len);
err_free_q:
	if (qfree) {
		if (qfree & 4) q_pool_destroy(tid, nthreads);
		if (qfree & 2) q_free(q);
		if (qfree & 1) free(q);
	}
err_free_tid:
	free(tid);
err_free_stree:
	mtree_free(&stree);
	if (err) errno = err;
	return rc;
}
